# -*- coding: utf-8 -*-

"""Unit test package for porerefiner."""

# from hypothesis_fspaths import fspaths

#safe_paths = lambda: fspaths().filter(lambda x: isinstance(x, str) or hasattr(x, '__fspath__'))

import asyncio
import tempfile

from unittest import TestCase

from collections import namedtuple
from datetime import datetime
from hypothesis.strategies import *
from pathlib import Path
import os
import namesgenerator

from porerefiner import models, jobs

from peewee import SqliteDatabase

from porerefiner.protocols.porerefiner.rpc import porerefiner_pb2 as messages

the_present = datetime.now()

# SQLite can't accept a 32-bit integer
sql_ints = lambda: integers(min_value=-2**16, max_value=2**16)

def _run(task):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(task)

@composite
def paths(draw, under="", min_deep=2, pathlib_only=False):
    r = [just(under)]
    for _ in range(min_deep):
        r.append(text(min_size=1, max_size=255))
    p = builds(Path, *r)
    if pathlib_only:
        return draw(p)
    return draw(one_of(p, p.map(str)))

@composite
def files(draw, real=False):
    if real:
        path = builds(Path, builds(lambda: tempfile.NamedTemporaryFile().name))
    else:
        path = paths(pathlib_only=True)
    return draw(builds(models.File,
                       pk=sql_ints(),
                       path=path))


@composite
def samples(draw):
    sid = text(min_size=7, max_size=10)
    acc = text(min_size=5, max_size=10)
    bar = text(min_size=10, max_size=10)
    org = text(min_size=12, max_size=12)
    ext = text(min_size=10, max_size=10)
    com = text()
    use = emails()
    return draw(builds(messages.SampleSheet.Sample,
                       sample_id=sid,
                       accession=acc,
                       barcode_id=bar,
                       organism=org,
                       extraction_kit=ext,
                       comment=com,
                       user=use))

@composite
def samplesheets(draw):
    ver = just('1.0.1')
    dat = datetimes()
    lib = text(min_size=12, max_size=12)
    seq = text(min_size=12, max_size=12)
    bar = text(min_size=12, max_size=12)
    sam = lists(samples(), min_size=1, max_size=12)
    ss = draw(builds(messages.SampleSheet,
                     porerefiner_ver=ver,
                     library_id=lib,
                     sequencing_kit=seq,
                     barcode_kit=bar,
                     samples=sam))
    ss.date.FromDatetime(draw(dat))
    return ss


@composite
def flowcells(draw, path=None):
    pk  = sql_ints()
    cid = text()
    cname = text()
    cty = text()
    if not path:
        pat = paths(min_deep=2, under="/data")
    else:
        pat = just(Path(path))
    return draw(builds(models.Flowcell,
                       #pk=pk,
                       consumable_id=cid,
                       consumable_name=cname,
                       consumable_type=cty,
                       path=pat))

@composite
def runs(draw, sheet=True):
    if sheet:
        ss = samplesheets()
    else:
        ss = just(None)
    path = draw(paths(pathlib_only=True))
    return draw(builds(models.Run,
                       #pk=sql_ints(),
                       # flowcell=flowcells(path=path.parent),
                       _sample_sheets=ss,
                       name=text(),
                       library_id=text(),
                       run_id=text(),
                       started=datetimes(max_value=the_present),
                       ended=datetimes(min_value=the_present),
                       status=sampled_from([status[0] for status in models.Run.statuses]),
                       path=just(path),
                       basecalling_model=sampled_from([model[0] for model in models.Run.basecallers])))

Event = namedtuple('Event', ('src_path', 'is_directory'))

@composite
def fsevents(draw, min_deep=4):
    return draw(builds(Event,
                       src_path=paths(min_deep=min_deep, pathlib_only=True),
                       is_directory=booleans()))

def random_name_subclass(of=object, **classdef):
    classdef['__module__'] = __name__
    new_typename = namesgenerator.get_random_name(sep=' ').title().replace(' ', '') + of.__name__
    new_type = type(new_typename, (of,), classdef)
    globals()[new_typename] = new_type
    return new_type

@composite
def submitters(draw, subclass_of=jobs.submitters.Submitter):
    return draw(builds(random_name_subclass(of=subclass_of,
                                            test_noop=lambda: None,
                                            reroot_path=lambda: None,
                                            begin_job=lambda: "",
                                            poll_job=lambda: "",
                                            closeout_job=lambda: None)))

@composite
def jobs(draw, subclass_of=jobs.RunJob, classdef=dict(setup=lambda *a, **k: None, collect=lambda *a, **k: None)):
    return draw(builds(random_name_subclass(of=subclass_of, **classdef),
                       submitter=submitters()))

class TestBase(TestCase):

    def setUp(self):
        self.db = SqliteDatabase(":memory:", pragmas={'foreign_keys':1}, autoconnect=False)
        self.db.bind(models.REGISTRY, bind_refs=False, bind_backrefs=False)
        self.db.connect()
        self.db.create_tables(models.REGISTRY)

    def tearDown(self):
        #[cls.delete().where(True).execute() for cls in models.REGISTRY]
        self.db.drop_tables(models.REGISTRY)
        self.db.close()

from functools import wraps

def with_database(func):
    @wraps(func)
    def wrapped_test_function(*a, **k):
        db = SqliteDatabase(":memory:", pragmas={'foreign_keys':1}, autoconnect=False)
        db.bind(models.REGISTRY, bind_refs=False, bind_backrefs=False)
        db.connect()
        db.create_tables(models.REGISTRY)
        try:
            return func(*a, **k)
        finally:
            db.drop_tables(models.REGISTRY)
            db.close()
    return wrapped_test_function

if __name__ == '__main__':
    symbol = None
    for symbol in locals().values():
        if hasattr(symbol, 'is_hypothesis_strategy_function'):
            try:
                print(symbol, symbol().example())
            except:
                pass



from abc import ABCMeta, abstractmethod, ABC
from collections import namedtuple
from dataclasses import dataclass

from .submitters import Submitter
from porerefiner.models import Job, File, Run
from porerefiner.cli_utils import render_dataclass, Email, Url, PathStr
from pathlib import Path
from typing import Union, Tuple

import logging
import pkgutil

log = logging.getLogger('porerefiner.job.registry')

async def poll_active_job(job):
    await job.job_state.submitter._poll(job)
    return 1

async def submit_job(job):
    job.job_state.submitter._submit(job)
    return 1

async def complete_job(job):
    job.job_state.submitter._close(job)
    return 1

async def poll_jobs():
    jobs_polled = 0
    jobs_submitted = 0
    jobs_collected = 0
    for job in Job.select().where(Job.status == 'READY'):
        jobs_submitted += await submit_job(job)
        jobs_polled += 1
    for job in Job.select().where(Job.status == 'RUNNING'):
        jobs_collected += await poll_active_job(job)
        jobs_polled += 1
    return jobs_polled, jobs_submitted, jobs_collected

JOBS = namedtuple('JOBS', ('FILES', 'RUNS'), defaults=([], []))() #configured (reified) job instances

REGISTRY = {} # available job classes

class _MetaRegistry(type):

    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        if cls not in REGISTRY:
            REGISTRY[name] = cls
        return cls

    def __call__(cls, *args, **kwargs):
        try:
            the_instance = super().__call__(*args, **kwargs)
        except TypeError as e:
            raise TypeError(f"Bad config: {cls.__name__}: {e} ({args}, {kwargs})") from e
        if isinstance(the_instance, FileJob):
            JOBS.FILES.append(the_instance)
            log.getChild('files').debug(cls.__name__)
        if isinstance(the_instance, RunJob):
            JOBS.RUNS.append(the_instance)
            log.getChild('runs').debug(cls.__name__)
        return the_instance

class RegisteringABCMeta(ABCMeta, _MetaRegistry):
    pass

@dataclass
class AbstractJob(metaclass=RegisteringABCMeta):
    submitter: Submitter

    @classmethod
    def get_configurable_options(cls):
        "Enumerate configurable options and value type as a guide to configuration."
        return render_dataclass(cls)


class FileJob(AbstractJob):

    @abstractmethod
    def setup(self, run: Run, file: File, datadir: Path, remotedir: Path) -> Union[str, Tuple[str, dict]]:
        pass

    @abstractmethod
    def collect(self, run: Run, file: File, datadir: Path, pid: Union[str, int]) -> None:
        pass

class RunJob(AbstractJob):

    @abstractmethod
    def setup(self, run: Run, datadir: Path, remotedir: Path) -> Union[str, Tuple[str, dict]]:
        pass

    @abstractmethod
    def collect(self, run: Run, datadir: Path, pid: Union[str, int]) -> None:
        pass



for loader, module_name, is_pkg in  pkgutil.walk_packages(__path__):
    _module = loader.find_module(module_name).load_module(module_name)
    # globals()[module_name] = _module



from tests import _run, fsevents, files

from unittest import TestCase, skip
from unittest.mock import Mock, patch
from mock import AsyncMock
from tempfile import NamedTemporaryFile

from porerefiner.fsevents import PoreRefinerFSEventHandler as Handler, end_file, end_run, register_new_run

from hypothesis import given
import hypothesis.strategies as strat

class TestPoreRefinerFSEventsHandler(TestCase):

    @skip('no flowcell support anymore')
    @patch('porerefiner.fsevents.register_new_flowcell', new_callable=AsyncMock)
    # @patch('porerefiner.fsevents.Flowcell')
    @given(event=fsevents().filter(lambda e: e.is_directory))
    def test_on_created_flowcell(self, event, reg):
        # flow.get_or_create.return_value = (flow, True)
        _run(Handler(event.src_path.parent).on_created(event))
        reg.assert_called()

    @skip("can't debug test")
    @given(event=fsevents().filter(lambda e: e.is_directory))
    def test_on_created_run(self, event):
        with patch('porerefiner.fsevents.register_new_run', new_callable=AsyncMock) as reg:
            with patch('porerefiner.fsevents.Run') as run:
                #self.assertEqual(len(event.src_path.parts), 3)
                run.get_or_create.return_value = (run, True)
                _run(Handler(event.src_path.parent.parent.parent).on_created(event))
                run.save.assert_called()

    @skip("can't debug test")
    # @patch('porerefiner.fsevents.Flowcell')
    @given(event=fsevents().filter(lambda e: not e.is_directory))
    def test_on_created_file(self, event):
        with patch('porerefiner.fsevents.Run') as run:
            with patch('porerefiner.fsevents.File') as file:
                run.get_or_create.return_value = (run, False)
                _run(Handler(event.src_path.parent.parent.parent.parent).on_created(event))
                file.create.assert_called()

class TestFSEventsPollingFunctions(TestCase):

    #@given(file=files(real=True))
    def test_end_file(self, file=Mock()):
        with NamedTemporaryFile() as tfile:
            file.path = tfile.name
            _run(end_file(file))
            file.save.assert_called()




from dataclasses import dataclass
from porerefiner.jobs.submitters import Submitter
from asyncio import subprocess


@dataclass
class LocalSubmitter(Submitter):

    async def test_noop(self):
        pass

    def reroot_path(self, path):
        return path #no op, no need to reroot paths

    async def begin_job(self, command, datadir, remotedir, environment_hints={}):
        self.process = await subprocess.create_subprocess_shell(command)

    async def poll_job(self, job):
        await self.process.wait()
        if self.process.returncode:
            return 'FAILED'
        return 'DONE'

    def closeout_job(self, job, datadir, remotedir):
        pass


